from typing import List

from marker.schema.page import Page


def get_pdf_toc(doc, max_depth=15):
    toc = doc.get_toc(max_depth=max_depth)
    toc_list = []
    for item in toc:
        list_item = {
            "title": item.title,
            "level": item.level,
            "page": item.page_index,
        }
        toc_list.append(list_item)
    return toc_list


def compute_toc(pages: List[Page]):
    toc = []
    for page in pages:
        for block in page.blocks:
            if block.block_type in ["Title", "Section-header"]:
                toc.append({
                    "title": block.prelim_text,
                    "level": block.heading_level,
                    "page": page.pnum
                })
    return toc



from typing import Optional, List, Dict, Literal

from dotenv import find_dotenv
from pydantic import computed_field
from pydantic_settings import BaseSettings
import torch
import os


class Settings(BaseSettings):
    # General
    TORCH_DEVICE: Optional[str] = None # Note: MPS device does not work for text detection, and will default to CPU
    IMAGE_DPI: int = 96 # DPI to render images pulled from pdf at
    EXTRACT_IMAGES: bool = True # Extract images from pdfs and save them
    PAGINATE_OUTPUT: bool = False # Paginate output markdown
    BASE_DIR: str = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    FLATTEN_PDF: bool = True # Pull form field values into the PDF before converting to markdown

    @computed_field
    @property
    def TORCH_DEVICE_MODEL(self) -> str:
        if self.TORCH_DEVICE is not None:
            return self.TORCH_DEVICE

        if torch.cuda.is_available():
            return "cuda"

        if torch.backends.mps.is_available():
            return "mps"

        return "cpu"

    DEFAULT_LANG: str = "English" # Default language we assume files to be in, should be one of the keys in TESSERACT_LANGUAGES

    SUPPORTED_FILETYPES: Dict = {
        "application/pdf": "pdf",
    }

    # Text extraction
    PDFTEXT_CPU_WORKERS: int = 4 # How many CPU workers to use for pdf text extraction

    # Text line Detection
    DETECTOR_BATCH_SIZE: Optional[int] = None # Defaults to 6 for CPU, 12 otherwise
    SURYA_DETECTOR_DPI: int = 96
    DETECTOR_POSTPROCESSING_CPU_WORKERS: int = 4

    # OCR
    INVALID_CHARS: List[str] = [chr(0xfffd), "�"]
    OCR_ENGINE: Optional[Literal["surya", "ocrmypdf"]] = "surya" # Which OCR engine to use, either "surya" or "ocrmypdf".  Defaults to "ocrmypdf" on CPU, "surya" on GPU.
    OCR_ALL_PAGES: bool = False # Run OCR on every page even if text can be extracted

    ## Surya
    SURYA_OCR_DPI: int = 192
    RECOGNITION_BATCH_SIZE: Optional[int] = None # Batch size for surya OCR defaults to 64 for cuda, 32 otherwise

    ## Tesseract
    OCR_PARALLEL_WORKERS: int = 2 # How many CPU workers to use for OCR
    TESSERACT_TIMEOUT: int = 20 # When to give up on OCR
    TESSDATA_PREFIX: str = ""

    # Texify model
    TEXIFY_MODEL_MAX: int = 384 # Max inference length for texify
    TEXIFY_TOKEN_BUFFER: int = 256 # Number of tokens to buffer above max for texify
    TEXIFY_DPI: int = 96 # DPI to render images at
    TEXIFY_BATCH_SIZE: Optional[int] = None # Defaults to 6 for cuda, 12 otherwise
    TEXIFY_MODEL_NAME: str = "vikp/texify"

    # Layout model
    SURYA_LAYOUT_DPI: int = 96
    BAD_SPAN_TYPES: List[str] = ["Page-footer", "Page-header", "Picture"] # You can add "Caption" and "Footnote" here to get rid of those elements - this just removes the text, not the image in case of Picture
    LAYOUT_MODEL_CHECKPOINT: str = "vikp/surya_layout3"
    BBOX_INTERSECTION_THRESH: float = 0.7 # How much the layout and pdf bboxes need to overlap to be the same
    TABLE_INTERSECTION_THRESH: float = 0.7
    LAYOUT_BATCH_SIZE: Optional[int] = None # Defaults to 12 for cuda, 6 otherwise
    DEFAULT_BLOCK_TYPE: str = "Text"

    # Ordering model
    SURYA_ORDER_DPI: int = 96
    ORDER_BATCH_SIZE: Optional[int] = None  # Defaults to 12 for cuda, 6 otherwise
    ORDER_MAX_BBOXES: int = 255

    # Table models
    SURYA_TABLE_DPI: int = 192
    TABLE_REC_BATCH_SIZE: Optional[int] = None

    # Headings
    HEADING_LEVEL_COUNT: int = 4
    HEADING_MERGE_THRESHOLD: float = .25
    HEADING_DEFAULT_LEVEL: int = 2

    # Output
    PAGE_SEPARATOR: str = "-" * 48 + "\n\n"

    # Debug
    DEBUG_DATA_FOLDER: str = os.path.join(BASE_DIR, "debug_data")
    DEBUG: bool = False
    FONT_DIR: str = os.path.join(BASE_DIR, "static", "fonts")
    DEBUG_RENDER_FONT: str = os.path.join(FONT_DIR, "GoNotoCurrent-Regular.ttf")
    FONT_DL_BASE: str = "https://github.com/satbyy/go-noto-universal/releases/download/v7.0"

    @computed_field
    @property
    def CUDA(self) -> bool:
        return "cuda" in self.TORCH_DEVICE_MODEL

    @computed_field
    @property
    def MODEL_DTYPE(self) -> torch.dtype:
        if self.TORCH_DEVICE_MODEL == "cuda":
            return torch.bfloat16
        else:
            return torch.float32

    @computed_field
    @property
    def TEXIFY_DTYPE(self) -> torch.dtype:
        return torch.float32 if self.TORCH_DEVICE_MODEL == "cpu" else torch.float16

    class Config:
        env_file = find_dotenv("local.env")
        extra = "ignore"


settings = Settings()


from tqdm import tqdm
from pypdfium2 import PdfDocument
from tabled.assignment import assign_rows_columns
from tabled.formats import formatter
from tabled.inference.detection import merge_tables

from surya.input.pdflines import get_page_text_lines
from tabled.inference.recognition import get_cells, recognize_tables

from marker.pdf.images import render_image
from marker.schema.bbox import rescale_bbox
from marker.schema.block import Line, Span, Block
from marker.schema.page import Page
from typing import List
from marker.ocr.recognition import get_batch_size as get_ocr_batch_size
from marker.ocr.detection import get_batch_size as get_detector_batch_size

from marker.settings import settings


def get_batch_size():
    if settings.TABLE_REC_BATCH_SIZE is not None:
        return settings.TABLE_REC_BATCH_SIZE
    elif settings.TORCH_DEVICE_MODEL == "cuda":
        return 6
    return 6


def get_table_boxes(pages: List[Page], doc: PdfDocument, fname):
    table_imgs = []
    table_counts = []
    table_bboxes = []
    img_sizes = []
    pnums = []

    for page_idx, page in enumerate(pages):
        # The bbox for the entire table
        bbox = [b.bbox for b in page.layout.bboxes if b.label == "Table"]
        highres_img = render_image(doc[page_idx], dpi=settings.SURYA_TABLE_DPI)

        page_table_imgs = []
        page_bboxes = []

        # Merge tables that are next to each other
        bbox = merge_tables(bbox)
        bbox = list(filter(lambda b: b[3] - b[1] > 10 and b[2] - b[0] > 10, bbox))

        if len(bbox) == 0:
            table_counts.append(0)
            img_sizes.append(None)
            pnums.append(page.pnum)
            continue

        # Number of tables per page
        table_counts.append(len(bbox))
        img_sizes.append(highres_img.size)
        pnums.append(page.pnum)

        for bb in bbox:
            highres_bb = rescale_bbox(page.layout.image_bbox, [0, 0, highres_img.size[0], highres_img.size[1]], bb)
            page_table_imgs.append(highres_img.crop(highres_bb))
            page_bboxes.append(highres_bb)

        table_imgs.extend(page_table_imgs)
        table_bboxes.extend(page_bboxes)

    # The page number in doc and in the original document are not the same
    # Doc has had pages removed from the start to align to start_page
    # This corrects for that
    doc_idxs = [pnum for pnum, tc in zip(pnums, table_counts) if tc > 0]
    table_idxs = [i for i, tc in enumerate(table_counts) if tc > 0]
    sel_text_lines = get_page_text_lines(
        fname,
        doc_idxs,
        [hr for i, hr in enumerate(img_sizes) if i in table_idxs],
        # Add flatten pdf here
    )
    text_lines = []
    out_img_sizes = []
    for i in range(len(table_counts)):
        if i in table_idxs:
            page_ocred = pages[i].ocr_method is not None
            if page_ocred:
                # This will force re-detection of cells if the page was ocred (the text lines are not accurate)
                text_lines.extend([None] * table_counts[i])
            else:
                text_lines.extend([sel_text_lines.pop(0)] * table_counts[i])
            out_img_sizes.extend([img_sizes[i]] * table_counts[i])

    assert len(table_imgs) == len(table_bboxes) == len(text_lines) == len(out_img_sizes)
    assert sum(table_counts) == len(table_imgs)

    return table_imgs, table_bboxes, table_counts, text_lines, out_img_sizes


def format_tables(pages: List[Page], doc: PdfDocument, fname: str, detection_model, table_rec_model, ocr_model):
    det_models = [detection_model, detection_model.processor]
    rec_models = [table_rec_model, table_rec_model.processor, ocr_model, ocr_model.processor]

    # Don't look at table cell detection tqdm output
    tqdm.disable = True
    table_imgs, table_boxes, table_counts, table_text_lines, img_sizes = get_table_boxes(pages, doc, fname)
    cells, needs_ocr = get_cells(table_imgs, table_boxes, img_sizes, table_text_lines, det_models, detect_boxes=settings.OCR_ALL_PAGES, detector_batch_size=get_detector_batch_size())
    tqdm.disable = False

    # This will redo OCR if OCR is forced, since we need to redetect bounding boxes, etc.
    table_rec = recognize_tables(table_imgs, cells, needs_ocr, rec_models, table_rec_batch_size=get_batch_size(), ocr_batch_size=get_ocr_batch_size())
    cells = [assign_rows_columns(tr, im_size) for tr, im_size in zip(table_rec, img_sizes)]
    table_md = [formatter("markdown", cell)[0] for cell in cells]

    table_count = 0
    for page_idx, page in enumerate(pages):
        page_table_count = table_counts[page_idx]
        if page_table_count == 0:
            continue

        table_insert_points = {}
        blocks_to_remove = set()
        pnum = page.pnum
        highres_size = img_sizes[table_count]
        page_table_boxes = table_boxes[table_count:table_count + page_table_count]

        for table_idx, table_box in enumerate(page_table_boxes):
            lowres_table_box = rescale_bbox([0, 0, highres_size[0], highres_size[1]], page.bbox, table_box)

            for block_idx, block in enumerate(page.blocks):
                intersect_pct = block.intersection_pct(lowres_table_box)
                if intersect_pct > settings.TABLE_INTERSECTION_THRESH and block.block_type == "Table":
                    if table_idx not in table_insert_points:
                        table_insert_points[table_idx] = max(0, block_idx - len(blocks_to_remove)) # Where to insert the new table
                    blocks_to_remove.add(block_idx)

        new_page_blocks = []
        for block_idx, block in enumerate(page.blocks):
            if block_idx in blocks_to_remove:
                continue
            new_page_blocks.append(block)

        for table_idx, table_box in enumerate(page_table_boxes):
            if table_idx not in table_insert_points:
                table_count += 1
                continue

            markdown = table_md[table_count]
            table_block = Block(
                bbox=table_box,
                block_type="Table",
                pnum=pnum,
                lines=[Line(
                    bbox=table_box,
                    spans=[Span(
                        bbox=table_box,
                        span_id=f"{table_idx}_table",
                        font="Table",
                        font_size=0,
                        font_weight=0,
                        block_type="Table",
                        text=markdown
                    )]
                )]
            )
            insert_point = table_insert_points[table_idx]
            insert_point = min(insert_point, len(new_page_blocks))
            new_page_blocks.insert(insert_point, table_block)
            table_count += 1
        page.blocks = new_page_blocks
    return table_count


from collections import defaultdict, Counter
from typing import List

from surya.layout import batch_layout_detection

from marker.pdf.images import render_image
from marker.schema.bbox import rescale_bbox
from marker.schema.block import bbox_from_lines
from marker.schema.page import Page
from marker.settings import settings


def get_batch_size():
    if settings.LAYOUT_BATCH_SIZE is not None:
        return settings.LAYOUT_BATCH_SIZE
    elif settings.TORCH_DEVICE_MODEL == "cuda":
        return 6
    return 6


def surya_layout(images: list, pages: List[Page], layout_model, batch_multiplier=1):
    text_detection_results = [p.text_lines for p in pages]

    processor = layout_model.processor
    layout_results = batch_layout_detection(images, layout_model, processor, detection_results=text_detection_results, batch_size=int(get_batch_size() * batch_multiplier))
    for page, layout_result in zip(pages, layout_results):
        page.layout = layout_result


def annotate_block_types(pages: List[Page]):
    for page in pages:
        max_intersections = {}
        for i, block in enumerate(page.blocks):
            for j, layout_block in enumerate(page.layout.bboxes):
                layout_bbox = layout_block.bbox
                layout_bbox = rescale_bbox(page.layout.image_bbox, page.bbox, layout_bbox)
                intersection_pct = block.intersection_pct(layout_bbox)
                if i not in max_intersections:
                    max_intersections[i] = (intersection_pct, j)
                elif intersection_pct > max_intersections[i][0]:
                    max_intersections[i] = (intersection_pct, j)

        for i, block in enumerate(page.blocks):
            block = page.blocks[i]
            block_type = None
            if i in max_intersections and max_intersections[i][0] > 0.0:
                j = max_intersections[i][1]
                block_type = page.layout.bboxes[j].label
            block.block_type = block_type

        # Smarter block layout assignment - first assign same as closest block
        # Next, fall back to text
        for i, block in enumerate(page.blocks):
            if block.block_type is not None:
                continue
            min_dist = None
            min_dist_idx = None
            for j, block2 in enumerate(page.blocks):
                if j == i or block2.block_type is None:
                    continue
                dist = block.distance(block2.bbox)
                if min_dist_idx is None or dist < min_dist:
                    min_dist = dist
                    min_dist_idx = j
                for line in block2.lines:
                    dist = block.distance(line.bbox)
                    if dist < min_dist:
                        min_dist = dist
                        min_dist_idx = j

            if min_dist_idx is not None:
                block.block_type = page.blocks[min_dist_idx].block_type

        for i, block in enumerate(page.blocks):
            if block.block_type is None:
                block.block_type = settings.DEFAULT_BLOCK_TYPE

        def get_layout_label(block_labels: List[str]):
            counter = Counter(block_labels)
            return counter.most_common(1)[0][0]

        def generate_block(block, block_labels):
            block.bbox = bbox_from_lines(block.lines)
            block.block_type = get_layout_label(block_labels)
            return block

        # Merge blocks together, preserving pdf order
        curr_layout_idx = None
        curr_layout_block = None
        curr_block_labels = []
        new_blocks = []
        for i in range(len(page.blocks)):
            if i not in max_intersections or max_intersections[i][0] == 0:
                if curr_layout_block is not None:
                    new_blocks.append(generate_block(curr_layout_block, curr_block_labels))
                curr_layout_block = None
                curr_layout_idx = None
                curr_block_labels = []
                new_blocks.append(page.blocks[i])
            elif max_intersections[i][1] != curr_layout_idx:
                if curr_layout_block is not None:
                    new_blocks.append(generate_block(curr_layout_block, curr_block_labels))
                curr_layout_block = page.blocks[i].copy()
                curr_layout_idx = max_intersections[i][1]
                curr_block_labels = [page.blocks[i].block_type]
            else:
                curr_layout_block.lines.extend(page.blocks[i].lines)
                curr_block_labels.append(page.blocks[i].block_type)

        if curr_layout_block is not None:
            new_blocks.append(generate_block(curr_layout_block, curr_block_labels))

        page.blocks = new_blocks



import argparse
import asyncio
import os

import requests
import uvicorn
from pydantic import BaseModel, Field
from starlette.responses import HTMLResponse

os.environ["PDFTEXT_CPU_WORKERS"] = "1"

import base64
from contextlib import asynccontextmanager
from typing import Optional, Annotated
import io

from fastapi import FastAPI, Body
from marker.convert import convert_single_pdf
from marker.models import load_all_models

app_data = {}

@asynccontextmanager
async def lifespan(app: FastAPI):
    if app.state.LOCAL:
        app_data["models"] = load_all_models()

    yield

    if "models" in app_data:
        del app_data["models"]


app = FastAPI(lifespan=lifespan)

@app.get("/")
async def root():
    return HTMLResponse(
"""
<h1>Marker API</h1>
<ul>
    <li><a href="/docs">API Documentation</a></li>
    <li><a href="/marker">Run marker (post request only)</a></li>
</ul>
"""
    )


class CommonParams(BaseModel):
    filepath: Annotated[
        str,
        Field(description="The path to the PDF file to convert.")
    ]
    max_pages: Annotated[
        Optional[int],
        Field(description="The maximum number of pages in the document to convert.", example=None)
    ] = None
    langs: Annotated[
        Optional[str],
        Field(description="The optional languages to use if OCR is needed, comma separated.  Must be either the names or codes from from https://github.com/VikParuchuri/surya/blob/master/surya/languages.py.", example=None)
    ] = None
    force_ocr: Annotated[
        bool,
        Field(description="Force OCR on all pages of the PDF.  Defaults to False.  This can lead to worse results if you have good text in your PDFs (which is true in most cases).")
    ] = False
    paginate: Annotated[
        bool,
        Field(description="Whether to paginate the output.  Defaults to False.  If set to True, each page of the output will be separated by a horizontal rule that contains the page number (2 newlines, {PAGE_NUMBER}, 48 - characters, 2 newlines).")
    ] = False
    extract_images: Annotated[
        bool,
        Field(description="Whether to extract images from the PDF.  Defaults to True.  If set to False, no images will be extracted from the PDF.")
    ] = True


@app.post("/marker")
async def convert_pdf(
    params: CommonParams
):
    if app.state.LOCAL:
        print(f"Converting {params.filepath} locally.")
        assert all([
            params.extract_images is True,
            params.paginate is False,
        ]), "Local conversion API does not support image extraction or pagination."
        return await convert_pdf_local(params)
    else:
        print(f"Converting {params.filepath} using the Datalab API.")
        return await convert_pdf_remote(params)


async def convert_pdf_remote(params: CommonParams):
    with open(params.filepath, "rb") as f:
        filedata = f.read()

    filename = os.path.basename(params.filepath)
    form_data = {
        'file': (filename, filedata, 'application/pdf'),
        'max_pages': (None, params.max_pages),
        'langs': (None, params.langs),
        'force_ocr': (None, params.force_ocr),
        'paginate': (None, params.paginate),
        'extract_images': (None, params.extract_images),
    }

    headers = {"X-API-Key": app.state.API_KEY}

    response = requests.post(app.state.DATALAB_URL, files=form_data, headers=headers)
    data = response.json()

    check_url = data["request_check_url"]

    for i in range(300):
        await asyncio.sleep(2)
        response = requests.get(check_url, headers=headers)
        data = response.json()

        if data["status"] == "complete":
            break

    return data


async def convert_pdf_local(params: CommonParams):
    try:
        full_text, images, metadata = convert_single_pdf(
            params.filepath,
            app_data["models"],
            max_pages=params.max_pages,
            langs=params.langs,
            ocr_all_pages=params.force_ocr
        )
    except Exception as e:
        return {
            "success": False,
            "error": str(e),
        }

    encoded = {}
    for k, v in images.items():
        byte_stream = io.BytesIO()
        v.save(byte_stream, format="PNG")
        encoded[k] = base64.b64encode(byte_stream.getvalue()).decode("utf-8")

    return {
        "markdown": full_text,
        "images": encoded,
        "metadata": metadata,
        "success": True
    }


def main():
    parser = argparse.ArgumentParser(description='Convert PDFs to markdown.')
    parser.add_argument('--port', type=int, default=8000, help='Port to run the server on')
    parser.add_argument('--host', type=str, default="127.0.0.1", help='Host to run the server on')
    parser.add_argument('--api_key', type=str, default=None, help='API key for the Datalab API.  If not specified, API will run locally.')
    parser.add_argument("--datalab_url", type=str, default="https://api.datalab.to/api/v1/marker", help="The URL for the Datalab API")

    args = parser.parse_args()

    app.state.API_KEY = args.api_key
    app.state.LOCAL = args.api_key is None
    app.state.DATALAB_URL = args.datalab_url

    # Run the server
    uvicorn.run(
        app,
        host=args.host,
        port=args.port,
    )


if __name__ == "__main__":
    main()


import json
import math
import os
from typing import List

from marker.debug.render import render_on_image
from marker.schema.bbox import rescale_bbox
from marker.schema.page import Page
from marker.settings import settings
from PIL import Image


def draw_layout_page_debug_images(fname, pages: List[Page]):
    # Remove extension from doc name
    doc_base = os.path.basename(fname).rsplit(".", 1)[0]

    debug_folder = os.path.join(settings.DEBUG_DATA_FOLDER, doc_base)
    os.makedirs(debug_folder, exist_ok=True)
    for idx, page in enumerate(pages):
        img_size = (int(math.ceil(page.text_lines.image_bbox[2])), int(math.ceil(page.text_lines.image_bbox[3])))
        png_image = Image.new("RGB", img_size, color="white")

        line_bboxes = []
        line_text = []
        for block in page.blocks:
            for line in block.lines:
                line_bboxes.append(rescale_bbox(page.bbox, page.text_lines.image_bbox, line.bbox))
                line_text.append(line.prelim_text)

        render_on_image(line_bboxes, png_image, labels=line_text, color="black", draw_bbox=False)
        pdf_image = png_image.copy()

        line_bboxes = [line.bbox for line in page.text_lines.bboxes]
        render_on_image(line_bboxes, png_image, color="blue")

        layout_boxes = [rescale_bbox(page.layout.image_bbox, page.text_lines.image_bbox, box.bbox) for box in page.layout.bboxes]
        layout_labels = [box.label for box in page.layout.bboxes]

        render_on_image(layout_boxes, png_image, labels=layout_labels, color="red")

        order_labels = [str(i) for i in range(len(page.layout.bboxes))]
        render_on_image(layout_boxes, png_image, labels=order_labels, color="green", draw_bbox=False, label_offset=5)

        debug_file = os.path.join(debug_folder, f"layout_page_{idx}.png")
        png_image.save(debug_file)

        # PDF Image

        block_bboxes = [rescale_bbox(page.bbox, page.text_lines.image_bbox, block.bbox) for block in page.blocks]
        block_labels = [block.block_type for block in page.blocks]
        render_on_image(block_bboxes, pdf_image, labels=block_labels, color="red")

        block_order = [str(i) for i in range(len(page.blocks))]
        render_on_image(block_bboxes, pdf_image, labels=block_order, color="green", draw_bbox=False, label_offset=5)

        debug_file = os.path.join(debug_folder, f"pdf_page_{idx}.png")
        pdf_image.save(debug_file)


def draw_pdf_page_debug_images(fname, pages: List[Page]):
    # Remove extension from doc name
    doc_base = os.path.basename(fname).rsplit(".", 1)[0]

    debug_folder = os.path.join(settings.DEBUG_DATA_FOLDER, doc_base)
    os.makedirs(debug_folder, exist_ok=True)
    for idx, page in enumerate(pages):
        img_size = (int(math.ceil(page.text_lines.image_bbox[2])), int(math.ceil(page.text_lines.image_bbox[3])))
        png_image = Image.new("RGB", img_size, color="white")

        line_bboxes = []
        line_text = []
        for block in page.blocks:
            for line in block.lines:
                line_bboxes.append(rescale_bbox(page.bbox, page.text_lines.image_bbox, line.bbox))
                line_text.append(line.prelim_text)




def draw_page_debug_images(fname, pages: List[Page]):
    if not settings.DEBUG:
        return

    draw_layout_page_debug_images(fname, pages)
    draw_pdf_page_debug_images(fname, pages)



def dump_bbox_debug_data(fname, pages: List[Page]):
    if not settings.DEBUG:
        return

    # Remove extension from doc name
    doc_base = os.path.basename(fname).rsplit(".", 1)[0]

    debug_file = os.path.join(settings.DEBUG_DATA_FOLDER, f"{doc_base}_bbox.json")
    debug_data = []
    for idx, page_blocks in enumerate(pages):
        page_data = page_blocks.model_dump(exclude=["images", "layout", "text_lines"])
        page_data["layout"] = page_blocks.layout.model_dump(exclude=["segmentation_map"])
        page_data["text_lines"] = page_blocks.text_lines.model_dump(exclude=["heatmap", "affinity_map"])
        debug_data.append(page_data)

    with open(debug_file, "w+") as f:
        json.dump(debug_data, f)
    print(f"Dumped bbox debug data to {debug_file}")






from collections import defaultdict
from typing import List
import numpy as np
from sklearn.cluster import KMeans

from marker.settings import settings
from marker.schema.bbox import rescale_bbox
from marker.schema.block import bbox_from_lines
from marker.schema.page import Page


def split_heading_blocks(pages: List[Page]):
    # Heading lines can be combined into regular text blocks sometimes by pdftext
    # Split up heading lines into separate blocks properly
    for page in pages:
        page_heading_boxes = [b for b in page.layout.bboxes if b.label in ["Title", "Section-header"]]
        page_heading_boxes = [(rescale_bbox(page.layout.image_bbox, page.bbox, b.bbox), b.label) for b in page_heading_boxes]

        new_blocks = []
        for block_idx, block in enumerate(page.blocks):
            if block.block_type not in ["Text"]:
                new_blocks.append(block)
                continue

            heading_lines = []
            for line_idx, line in enumerate(block.lines):
                for (heading_box, label) in page_heading_boxes:
                    if line.intersection_pct(heading_box) > settings.BBOX_INTERSECTION_THRESH:
                        heading_lines.append((line_idx, label))
                        break

            if len(heading_lines) == 0:
                new_blocks.append(block)
                continue

            # Split up the block into separate blocks around headers
            start = 0
            for (heading_line, label) in heading_lines:
                if start < heading_line:
                    copied_block = block.copy()
                    copied_block.lines = block.lines[start:heading_line]
                    copied_block.bbox = bbox_from_lines(copied_block.lines)
                    new_blocks.append(copied_block)

                copied_block = block.copy()
                copied_block.lines = block.lines[heading_line:heading_line + 1]
                copied_block.block_type = label
                copied_block.bbox = bbox_from_lines(copied_block.lines)
                new_blocks.append(copied_block)

                start = heading_line + 1
                if start >= len(block.lines):
                    break

            # Add any remaining lines
            if start < len(block.lines):
                copied_block = block.copy()
                copied_block.lines = block.lines[start:]
                copied_block.bbox = bbox_from_lines(copied_block.lines)
                new_blocks.append(copied_block)

        page.blocks = new_blocks


def bucket_headings(line_heights, num_levels=settings.HEADING_LEVEL_COUNT):
    if len(line_heights) <= num_levels:
        return []

    data = np.asarray(line_heights).reshape(-1, 1)
    labels = KMeans(n_clusters=num_levels, random_state=0, n_init="auto").fit_predict(data)
    data_labels = np.concatenate([data, labels.reshape(-1, 1)], axis=1)
    data_labels = np.sort(data_labels, axis=0)

    cluster_means = {int(label): float(np.mean(data_labels[data_labels[:, 1] == label, 0])) for label in np.unique(labels)}
    label_max = None
    label_min = None
    heading_ranges = []
    prev_cluster = None
    for row in data_labels:
        value, label = row
        value = float(value)
        label = int(label)
        if prev_cluster is not None and label != prev_cluster:
            prev_cluster_mean = cluster_means[prev_cluster]
            cluster_mean = cluster_means[label]
            if cluster_mean * settings.HEADING_MERGE_THRESHOLD < prev_cluster_mean:
                heading_ranges.append((label_min, label_max))
                label_min = None
                label_max = None

        label_min = value if label_min is None else min(label_min, value)
        label_max = value if label_max is None else max(label_max, value)
        prev_cluster = label

    if label_min is not None:
        heading_ranges.append((label_min, label_max))

    heading_ranges = sorted(heading_ranges, reverse=True)

    return heading_ranges


def infer_heading_levels(pages: List[Page], height_tol=.99):
    all_line_heights = []
    for page in pages:
        for block in page.blocks:
            if block.block_type not in ["Title", "Section-header"]:
                continue

            all_line_heights.extend([l.height for l in block.lines])

    heading_ranges = bucket_headings(all_line_heights)

    for page in pages:
        for block in page.blocks:
            if block.block_type not in ["Title", "Section-header"]:
                continue

            block_heights = [l.height for l in block.lines]
            if len(block_heights) > 0:
                avg_height = sum(block_heights) / len(block_heights)
                for idx, (min_height, max_height) in enumerate(heading_ranges):
                    if avg_height >= min_height * height_tol:
                        block.heading_level = idx + 1
                        break

            if block.heading_level is None:
                block.heading_level = settings.HEADING_DEFAULT_LEVEL




import requests
from PIL import ImageDraw, ImageFont, Image

from marker.settings import settings
import os


def get_font_path() -> str:
    font_path = settings.DEBUG_RENDER_FONT

    if not os.path.exists(font_path):
        os.makedirs(os.path.dirname(font_path), exist_ok=True)
        font_dl_path = f"{settings.FONT_DL_BASE}/{os.path.basename(font_path)}"
        with requests.get(font_dl_path, stream=True) as r, open(font_path, 'wb') as f:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

    return font_path


def get_text_size(text, font):
    im = Image.new(mode="P", size=(0, 0))
    draw = ImageDraw.Draw(im)
    _, _, width, height = draw.textbbox((0, 0), text=text, font=font)
    return width, height


def render_on_image(bboxes, image, labels=None, label_offset=1, label_font_size=10, color: str | list='red', draw_bbox=True):
    draw = ImageDraw.Draw(image)
    font_path = get_font_path()
    label_font = ImageFont.truetype(font_path, label_font_size)

    for i, bbox in enumerate(bboxes):
        bbox = [int(p) for p in bbox]
        if draw_bbox:
            draw.rectangle(bbox, outline=color[i] if isinstance(color, list) else color, width=1)

        if labels is not None:
            label = labels[i]
            text_position = (
                bbox[0] + label_offset,
                bbox[1] + label_offset
            )
            text_size = get_text_size(label, label_font)
            if text_size[0] <= 0 or text_size[1] <= 0:
                continue
            box_position = (
                text_position[0],
                text_position[1],
                text_position[0] + text_size[0],
                text_position[1] + text_size[1]
            )
            draw.rectangle(box_position, fill="white")
            draw.text(
                text_position,
                label,
                fill=color[i] if isinstance(color, list) else color,
                font=label_font
            )

    return image